#  Copyright (c) 2022.
#  All rights reserved to the creator of the following script/program/app, please do not
#  use or distribute without prior authorization from the creator.
#  Creator: Antonio Manuel Nunes Goncalves
#  Email: amng835@gmail.com
#  LinkedIn: https://www.linkedin.com/in/antonio-manuel-goncalves-983926142/
#  Github: https://github.com/DEADSEC-SECURITY

# Built-In Imports

# 3rd-Party Imports

# Local Imports



#  Copyright (c) 2022.
#  All rights reserved to the creator of the following script/program/app, please do not
#  use or distribute without prior authorization from the creator.
#  Creator: Antonio Manuel Nunes Goncalves
#  Email: amng835@gmail.com
#  LinkedIn: https://www.linkedin.com/in/antonio-manuel-goncalves-983926142/
#  Github: https://github.com/DEADSEC-SECURITY

# Built-In Imports
import pathlib
import shutil
import unittest

# 3rd-Party Imports
from aws_cdk import aws_lambda as lambda_

# Local Imports
from aws_cdk_code_builder import Build

CURRENT_FILE = pathlib.Path(__file__).absolute()
CURRENT_PATH = CURRENT_FILE.parent
BUILD_PATH = CURRENT_PATH.joinpath('.build')
HASHES_PATH = BUILD_PATH.joinpath('.hashes')
CODE_FOLDER_NAME = 'test_code'
CODE_PATH = CURRENT_PATH.joinpath(CODE_FOLDER_NAME)
BUILD_CODE_PATH = BUILD_PATH.joinpath(CODE_FOLDER_NAME)
HASH_CODE_PATH = HASHES_PATH.joinpath(CODE_FOLDER_NAME)


def _make_file(path, content):
    with open(path, 'w+') as file:
        file.write(content)


class TestBuild(unittest.TestCase):
    @staticmethod
    def default():
        if BUILD_PATH.exists():
            shutil.rmtree(BUILD_PATH)
        if CODE_PATH.exists():
            shutil.rmtree(CODE_PATH)
        CODE_PATH.mkdir()
        CODE_PATH.joinpath('sub_folder').mkdir()

        _make_file(CODE_PATH.joinpath('requirements.txt'), 'popcorn-time')
        _make_file(CODE_PATH.joinpath('code.py'), 'print("Hello World")')
        _make_file(CODE_PATH.joinpath('sub_folder').joinpath('code2.py'), 'print("HI")')

    def test_build(self):
        self.default()
        package = Build(
            project_path=CODE_PATH,
            work_dir=CURRENT_PATH
        ).build()

        self.assertTrue(isinstance(package, lambda_.Code), 'Wrong return type for build method')
        self.assertTrue(BUILD_PATH.exists(), 'Build directory not found')
        self.assertTrue(HASHES_PATH.exists(), 'Hashes directory not found')
        self.assertTrue(BUILD_CODE_PATH.exists(), 'Code build directory not found')
        self.assertTrue(BUILD_CODE_PATH.joinpath('popcorntime').exists(), 'External libraries not installed properly')
        self.assertTrue(HASH_CODE_PATH.exists(), 'Code hash directory not found')
        self.assertTrue(HASH_CODE_PATH.joinpath('hash.txt').exists(), 'Code hash file not found')

    def test_build_with_cache(self):
        self.test_build()

        package = Build(
            project_path=CODE_PATH,
            work_dir=CURRENT_PATH
        ).build()

        self.assertTrue(isinstance(package, lambda_.Code), 'Wrong return type for build method')
        self.assertTrue(BUILD_PATH.exists(), 'Build directory not found')
        self.assertTrue(HASHES_PATH.exists(), 'Hashes directory not found')
        self.assertTrue(BUILD_CODE_PATH.exists(), 'Code build directory not found')
        self.assertTrue(BUILD_CODE_PATH.joinpath('popcorntime').exists(), 'External libraries not installed properly')
        self.assertTrue(HASH_CODE_PATH.exists(), 'Code hash directory not found')
        self.assertTrue(HASH_CODE_PATH.joinpath('hash.txt').exists(), 'Code hash file not found')



#  Copyright (c) 2022.
#  All rights reserved to the creator of the following script/program/app, please do not
#  use or distribute without prior authorization from the creator.
#  Creator: Antonio Manuel Nunes Goncalves
#  Email: amng835@gmail.com
#  LinkedIn: https://www.linkedin.com/in/antonio-manuel-goncalves-983926142/
#  Github: https://github.com/DEADSEC-SECURITY

# Built-In Imports
import os
import shutil
import hashlib
from typing import Union
from distutils import dir_util

# 3rd-Party Imports
from pathlib import Path
from aws_cdk import aws_lambda as lambda_


# Local Imports


def recursive_copy(src, dst):
    os.chdir(src)
    for item in os.listdir():

        if os.path.isfile(item):
            shutil.copy(item, dst)

        elif os.path.isdir(item):
            new_dst = os.path.join(dst, item)
            os.mkdir(new_dst)
            recursive_copy(os.path.abspath(item), new_dst)


class Build:
    project_path: Path
    work_dir: Path
    # Internal
    _build_dir: Path
    _project_build_dir: Path
    _req_file: Path

    def __init__(self, project_path: Union[str, Path], work_dir: Union[str, Path]):
        self.project_path = Path(project_path).resolve() if isinstance(project_path, str) else project_path
        self.work_dir = Path(work_dir).resolve() if isinstance(work_dir, str) else work_dir

        if not self.work_dir.is_dir():
            raise ValueError('Work dir needs to be a directory')
        if not self.project_path.is_dir():
            raise ValueError('Project path needs to be a directory')

        project_name = self.project_path.parts[-1]

        self._build_dir = self.work_dir.joinpath('.build')
        self._project_build_dir = self._build_dir.joinpath(project_name)

        if not self._build_dir.exists():
            self._build_dir.mkdir()

        if not self._project_build_dir.exists():
            self._project_build_dir.mkdir()

        self._req_file = self._project_build_dir.joinpath('requirements.txt')
        self._hash_cache_dir = self._build_dir.joinpath('.hashes')
        self._project_hash_dir = self._hash_cache_dir.joinpath(project_name)

        if not self._hash_cache_dir.exists():
            self._hash_cache_dir.mkdir()

        if not self._project_hash_dir.exists():
            self._project_hash_dir.mkdir()

        self._hash_file = self._project_hash_dir.joinpath('hash.txt')

    def _get_cached_hash(self):
        if not self._hash_file.exists():
            return False

        with open(self._hash_file, 'r') as file:
            return file.read()

    def _get_current_hash(self):
        with open(self._req_file, 'rb') as req:
            req_hash = hashlib.sha1(req.read()).hexdigest()
        return req_hash

    def _save_hash(self, hash_: str):
        with open(self._hash_file, 'w+') as file:
            file.write(hash_)

    def _generate_new_hash(self):
        self._save_hash(self._get_current_hash())

    def _reset_build_folder(self):
        if self._project_build_dir.exists():
            shutil.rmtree(self._project_build_dir)
        self._project_build_dir.mkdir()
        recursive_copy(self.project_path, self._project_build_dir)

    def _build(self):
        self._reset_build_folder()
        os.system(f'pip3 install -r {self._req_file} -t {self._project_build_dir}')
        self._generate_new_hash()
        return lambda_.Code.from_asset(self._project_build_dir.as_posix())

    def build(self):
        old_hash = self._get_cached_hash()

        if not old_hash:
            return self._build()

        if self._get_current_hash() != old_hash:
            return self._build()

        return lambda_.Code.from_asset(self._project_build_dir.as_posix())
